/*
** Der umstlpbare Wrfel nach Paul Schatz
** Copyright (C) 1999,2004 Franz Zahaurek <http://www.fzk.at>
**
**    This program is free software; you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation; either version 2 of the License, or
**    (at your option) any later version.
**
*/
/* $Id: wdisplists.c,v 1.11 2004/05/26 17:34:57 fzk Exp $ */

/* Erstellen der Display-Listen */

#ifndef OS_WIN32
  #include <config.h>
#endif
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#ifndef OS_WIN32
#include <sys/time.h>
#endif

/* #undef CHECKOGLERR */
#include "wuerfel.h"

void wAxisTransform (void);
void wCreateLines (void);
void wCreateFill  (void);
void wCreateLabels (void);
void wCreateVirt (void);
void wCreateAxis (void);
void wCreateTouchline (void);
void wCreateTimeform1 (void);
void wCreateTimeform2 (void);
void wCreateDiagonals (void);
void wCreateCenter (void);
void wDisplayGlMatrices  (void);
void wDisplayMatrix  (char *text, GLfloat *matrix16x16);

extern int    callFill, callLines, callVirt, callAxis, callTimeform, callTouchline;

#define OP(n,i) owgp->wgp.e[(n)-1].i
#define VON(n)  OP((n),x),OP((n),y),OP((n),z)
#define W       owgp->wgp.w
#define W2      owgp->wgp.w2
#define NACH(n) VON(n)

void
wCreateDisplayLists () {
    int listId;
    WG *aktupos;
    static int e = 1; /* ecke */
    static int z = 1;
    RPUNKT zwi, prevzwi;

/*     fprintf (stderr, "Neue DisplayLists options.oloid=%s, options.kuboid=%s\n", */
/* 	     options.oloid?"true":"false", options.kuboid?"true":"false"); */

    aktupos = owgp;
    /*
    ** Die Gitterdarstellung des Wuerfelguertels
    ** Beginnend mit listId=1 wird fuer jede berechneten Positionen eine
    ** glNewList() gebildet.  Diese glNewList() werden in der Variablen anzpos
    ** gezaehlt, die auch fuer die uebrigen Listen gilt/gelten sollte.
    **
    ** Das hier Gesagte, gilt auch fuer die folgenden Display-Lists, die jeweils
    ** eine Position darstellen.
    */
    listIdLines = listIdLinesStart = listId = 1;
    for (owgp = wg; owgp; owgp = owgp->nextwgp) {
        wErr(glNewList (listId, GL_COMPILE));

	/* Beginn Achsentransformation fuer Liniendarstellung des Wuerfelguertels */
	wErr(glPushMatrix());

	wAxisTransform();

	wCreateLines();
	wCreateLabels();

	wErr(glPopMatrix());

	wErr(glEndList());
        anzpos = listId;   /* anzpos gilt fuer alle Objekt-Listen */
        listId++;
    }

/*     fprintf (stderr, "Anzahl der Positionen in listIdLines =%d\n", */
/* 	     listId - listIdLines); */

    /*
    ** Die flaechige Darstellung des Wuerfelguertels
    ** Beginnend mit der hier aktuellen listId wird fuer jede berechneten Positionen
    ** eine glNewList () gebildet.
    */
    listIdFill = listIdFillStart = listId;
    for (owgp = wg; owgp; owgp = owgp->nextwgp) {
        wErr(glNewList (listId, GL_COMPILE));

	/* Beginn Achsentransformation fuer Flchendarstellung des Wuerfelguertels */
	wErr(glPushMatrix());

	wAxisTransform();

	wCreateFill();
	wCreateLabels();

	wErr(glPopMatrix());

	wErr(glEndList());
	listId++;
    }

/*     fprintf (stderr, "Anzahl der Positionen in listIdFill =%d\n", */
/* 	     listId - listIdFill); */

    /*
    ** Die Beruehrungslinien Timeform - Wuerfelguertel
    */
    listIdTouch = listIdTouchStart = listId;
    for (owgp = wg; owgp; owgp = owgp->nextwgp) {
        wErr(glNewList (listId, GL_COMPILE));

	wErr(glPushMatrix());

        wAxisTransform();
	wCreateTouchline();

	wErr(glPopMatrix());

	wErr(glEndList());
	listId++;
    }

/*     fprintf (stderr, "Anzahl der Positionen in listIdTouch =%d\n", */
/* 	     listId - listIdTouch); */

    /*
    ** Die 3 konstanten Diagonalen des Wuerfelguertels
    ** Verbindung Evolutions- und Involutionspunkt waere evtl. auch interessant
    */
    listIdDiagonals = listIdDiagonalsStart = listId;
    for (owgp = wg; owgp; owgp = owgp->nextwgp) {
        wErr(glNewList (listId, GL_COMPILE));

	wErr(glPushMatrix());

	wAxisTransform();
	wCreateDiagonals();

	wErr(glPopMatrix());

        wErr(glEndList());
        listId++;
    }

    /*
    ** Der Mittelpunkt des Wuerfelguertels
    */
    listIdCenter = listIdCenterStart = listId;
    for (owgp = wg; owgp; owgp = owgp->nextwgp) {
        wErr(glNewList (listId, GL_COMPILE));

	wErr(glPushMatrix());

	wAxisTransform();
	wCreateCenter();

	wErr(glPopMatrix());

        wErr(glEndList());
        listId++;
    }

    /*
    ** Die 6 gedachten/virtuellen Verbindungen der Scharniere des Wuerfelguertels
    ** mit Evolutions- und Involutionspunkt
    */
    listIdVirt = listIdVirtStart = listId;
    for (owgp = wg; owgp; owgp = owgp->nextwgp) {
        wErr(glNewList (listId, GL_COMPILE));

	wErr(glPushMatrix());

	wAxisTransform();
	wCreateVirt();

	wErr(glPopMatrix());

        wErr(glEndList());
        listId++;
    }

/*     fprintf (stderr, "Anzahl der Positionen in listIdVirt =%d\n", */
/* 	     listId - listIdVirt); */

    /*
    ** Die Koordinatenachsen
    */
    listIdAxis = listIdAxisStart = listId;
    for (owgp = wg; owgp; owgp = owgp->nextwgp) {
        wErr(glNewList (listId, GL_COMPILE));

	wErr(glPushMatrix());
	wAxisTransform();

	wCreateAxis();

	wErr(glPopMatrix());

        wErr(glEndList());
        listId++;
    }

/*     fprintf (stderr, "Anzahl der Positionen in listIdAxis =%d\n", */
/* 	     listId - listIdAxis ); */

    /*
    ** Die Zeitformen Oloid und Kuboid Teil A) die Erzeugende
    */
    listIdTime1 = listIdTime1Start = listId;
    wErr(glNewList (listId, GL_COMPILE));

    wCreateTimeform1();

    wErr(glEndList());
    listId++;

/*     fprintf (stderr, "Anzahl der Positionen in listIdTime1 =%d\n", */
/* 	     listId - listIdTime1); */

    /*
    ** Die Zeitformen Oloid und Kuboid Teil B) die Endpunkte der Erzeugenden
    */
    listIdTime2 = listIdTime2Start = listId;
    wErr(glNewList (listId, GL_COMPILE));

    wCreateTimeform2();

    wErr(glEndList());
    listId++;

/*     fprintf (stderr, "Anzahl der Positionen in listIdTime2 =%d\n", */
/* 	     listId - listIdTime2); */


    owgp = wg; // aktupos;
}

void
wAxisTransform () {
	/*
	** Achsentransformation so durchfuehren, dass das Oloid/Kuboid
	** erkennbar wird.
	*/
	if (options.axistrans == 2) {
	  wErr(glRotatef (GWINKEL(owgp->wgp.w2), 1., 0., 0.));
	  wErr(glRotatef (GWINKEL(owgp->wgp.w),  0., 1., 0.));
	}
	if (options.axistrans == 1) {
	  wErr(glRotatef (GWINKEL(owgp->wgp.w),  0., 1., 0.));
	}
/* 	if (options.mode == 4) { */
/* 	  fprintf (stderr, "."); */
/* 	  wErr(glTranslatef (owgp->wgp.e[2-1].x < 0. ? -owgp->wgp.e[2-1].x : 0., 0., 0.)); */
/* 	} */

/*	wDisplayGlMatrices (); */
}

void
wCreateLines () {
        wErr(glColor3f (1.0, 1.0, 1.0));
        wErr(glClearColor (0.7, 0.7, 0.0, 0.0));
        wErr(glLineWidth (1.0));
        wErr(glEnable (GL_LINE_SMOOTH));

        wErr(glBegin (GL_TRIANGLE_FAN));
          wErr(glVertex3f (VON(3)));
          wErr(glVertex3f (VON(1)));
          wErr(glVertex3f (VON(2)));
          wErr(glVertex3f (VON(4)));
          wErr(glVertex3f (VON(1)));
        wErr(glEnd ());
        wErr(glBegin (GL_TRIANGLE_FAN));
          wErr(glVertex3f (VON(7)));
          wErr(glVertex3f (VON(8)));
          wErr(glVertex3f (VON(3)));
          wErr(glVertex3f (VON(4)));
          wErr(glVertex3f (VON(8)));
        wErr(glEnd ());
        wErr(glBegin (GL_TRIANGLE_FAN));
          wErr(glVertex3f (VON(5)));
          wErr(glVertex3f (VON(6)));
          wErr(glVertex3f (VON(7)));
          wErr(glVertex3f (VON(8)));
          wErr(glVertex3f (VON(6)));
        wErr(glEnd ());
        wErr(glBegin (GL_TRIANGLE_FAN));
          wErr(glVertex3f (VON(13)));
          wErr(glVertex3f (VON(11)));
          wErr(glVertex3f (VON(12)));
          wErr(glVertex3f (VON(14)));
          wErr(glVertex3f (VON(11)));
        wErr(glEnd ());
        wErr(glBegin (GL_TRIANGLE_FAN));
          wErr(glVertex3f (VON(17)));
          wErr(glVertex3f (VON(18)));
          wErr(glVertex3f (VON(13)));
          wErr(glVertex3f (VON(14)));
          wErr(glVertex3f (VON(18)));
        wErr(glEnd ());
        wErr(glBegin (GL_TRIANGLE_FAN));
          wErr(glVertex3f (VON(15)));
          wErr(glVertex3f (VON(16)));
          wErr(glVertex3f (VON(17)));
          wErr(glVertex3f (VON(18)));
          wErr(glVertex3f (VON(16)));
        wErr(glEnd ());

        wErr(glPointSize(4.));
	wErr(glBegin (GL_POINTS));
          wErr(glVertex3f (VON(1)));
          wErr(glVertex3f (VON(5)));
        wErr(glEnd ());
}

void
wCreateFill () {
        wErr(glClearColor (0.7, 0.7, 0.0, 0.0));
        wErr(glBegin (GL_TRIANGLE_FAN));
          wErr(glVertex3f (VON(3))); wErr(glColor3f (VON(3)));
          wErr(glVertex3f (VON(1))); wErr(glColor3f (VON(1)));
          wErr(glVertex3f (VON(2))); wErr(glColor3f (VON(2)));
          wErr(glVertex3f (VON(4))); wErr(glColor3f (VON(4)));
          wErr(glVertex3f (VON(1))); wErr(glColor3f (VON(1)));
        wErr(glEnd ());
        wErr(glBegin (GL_TRIANGLE_FAN));
          wErr(glVertex3f (VON(7))); wErr(glColor3f (VON(7)));
          wErr(glVertex3f (VON(8))); wErr(glColor3f (VON(8)));
          wErr(glVertex3f (VON(3))); wErr(glColor3f (VON(3)));
          wErr(glVertex3f (VON(4))); wErr(glColor3f (VON(4)));
          wErr(glVertex3f (VON(8))); wErr(glColor3f (VON(8)));
        wErr(glEnd ());
        wErr(glBegin (GL_TRIANGLE_FAN));
          wErr(glVertex3f (VON(5))); wErr(glColor3f (VON(5)));
          wErr(glVertex3f (VON(6))); wErr(glColor3f (VON(6)));
          wErr(glVertex3f (VON(7))); wErr(glColor3f (VON(7)));
          wErr(glVertex3f (VON(8))); wErr(glColor3f (VON(8)));
          wErr(glVertex3f (VON(6))); wErr(glColor3f (VON(6)));
        wErr(glEnd ());
        wErr(glBegin (GL_TRIANGLE_FAN));
          wErr(glVertex3f (VON(13))); wErr(glColor3f (VON(3)));
          wErr(glVertex3f (VON(11))); wErr(glColor3f (VON(1)));
          wErr(glVertex3f (VON(12))); wErr(glColor3f (VON(2)));
          wErr(glVertex3f (VON(14))); wErr(glColor3f (VON(4)));
          wErr(glVertex3f (VON(11))); wErr(glColor3f (VON(1)));
        wErr(glEnd ());
        wErr(glBegin (GL_TRIANGLE_FAN));
          wErr(glVertex3f (VON(17))); wErr(glColor3f (VON(7)));
          wErr(glVertex3f (VON(18))); wErr(glColor3f (VON(8)));
          wErr(glVertex3f (VON(13))); wErr(glColor3f (VON(3)));
          wErr(glVertex3f (VON(14))); wErr(glColor3f (VON(4)));
          wErr(glVertex3f (VON(18))); wErr(glColor3f (VON(8)));
        wErr(glEnd ());
        wErr(glBegin (GL_TRIANGLE_FAN));
          wErr(glVertex3f (VON(15))); wErr(glColor3f (VON(5)));
          wErr(glVertex3f (VON(16))); wErr(glColor3f (VON(6)));
          wErr(glVertex3f (VON(17))); wErr(glColor3f (VON(7)));
          wErr(glVertex3f (VON(18))); wErr(glColor3f (VON(8)));
          wErr(glVertex3f (VON(16))); wErr(glColor3f (VON(6)));
        wErr(glEnd ());

        wErr(glPointSize(4.));
        wErr(glColor3f (1., 1., 1.));
        wErr(glBegin (GL_POINTS));
          wErr(glVertex3f (VON(1)));
          wErr(glVertex3f (VON(5)));
        wErr(glEnd ());
}


/* Labels in x-Richtung etwas nach rechts verschoben */
#define VONDIST(n)  OP((n),x)+.05,OP((n),y),OP((n),z)

void
wCreateLabels () {
  GLfloat params[4];

  wErr(glGetFloatv(GL_CURRENT_COLOR, params)); 
  wErr(glColor3f (0., 0., 0.)); /* Labels immer in Schwarz */
        if (options.labels) {
	  wErr(glRasterPos3f (VONDIST(1))); wDrawString ("1");
	  wErr(glRasterPos3f (VONDIST(2))); wDrawString ("2");
	  wErr(glRasterPos3f (VONDIST(3))); wDrawString ("3");
	  wErr(glRasterPos3f (VONDIST(4))); wDrawString ("4");

	  wErr(glRasterPos3f (VONDIST(5))); wDrawString ("5");
	  wErr(glRasterPos3f (VONDIST(6))); wDrawString ("6");
	  wErr(glRasterPos3f (VONDIST(7))); wDrawString ("7");
	  wErr(glRasterPos3f (VONDIST(8))); wDrawString ("8");

	  wErr(glRasterPos3f (VONDIST(17))); wDrawString ("17");
	  if (options.center) {
	    wErr(glRasterPos3f (VONDIST(22))); wDrawString ("M");
	  }
        }
  wErr(glColor4fv (params));
}

void
wCreateTouchline () {
  GLfloat params[4];

	/*
	** Beruehrungslinie zwischen Oloid/Kuboid und Wrfelgrtel
	** nur, wenn auch eines der beiden Objekte dargestellt wird
	*/
	wErr(glLineWidth (4.0));
	//        wErr(glEnable (GL_LINE_SMOOTH));
	wErr(glColor3f (1.,1.,1.));
        wErr(glClearColor (0.7, 0.7, 0.0, 0.0));
	wErr(glBegin (GL_LINES));

	if (options.oloid || options.beide) {
	  wErr(glVertex3f (VON(3))); 
	  wErr(glVertex3f (VON(17)));
	}
	if (options.kuboid || options.beide) {
	  wErr(glVertex3f (VON(3)));
	  wErr(glVertex3f (VON(7)));
	  wErr(glVertex3f (VON(7)));
	  wErr(glVertex3f (VON(5)));
	}
	wErr(glEnd ());

	/*
	** Radien zu den Enden der Berhrungslinie(n)
	** vielleicht doch nicht so anschaulich ?
	*/
	wErr(glLineWidth (1.0));
	wErr(glColor3f (.8,.8,.3));
        wErr(glClearColor (0.7, 0.7, 0.0, 0.0));
	wErr(glBegin (GL_LINES));

	if (options.oloid || options.beide) {
	  wErr(glVertex3f (VON(1))); 
	  wErr(glVertex3f (VON(3)));
	  wErr(glVertex3f (VON(17))); 
	  wErr(glVertex3f (VON(13)));
	  wErr(glVertex3f (VON(13)));
	  wErr(glVertex3f (VON(1)));
	}
	if (options.kuboid || options.beide) {
	  wErr(glVertex3f (VON(1))); 
	  wErr(glVertex3f (VON(3)));
	  wErr(glVertex3f (VON(5)));
	  wErr(glVertex3f (VON(1))); 
	}

	wErr(glEnd ());
}

/*
** Makros fuer die manuelle Achsentransformation x/y/z
** und zwar jeweils die aktuelle und die vorherige Position (previous)
*/
#define OPP(n,i) prevwgp->wgp.e[(n)-1].i
#define PREVVON(n) OPP((n),x),OPP((n),y),OPP((n),z)
#define WP       (prevwgp->wgp.w)
#define W2P      (prevwgp->wgp.w2)


void
wCreateTimeform1 () {
	WG *owgp;
	WG *actualwgp;
	WG *lastactualwgp;
	int pos;

	/*
	** Oloid / Kuboid Verbindungslinien jeweils einer Position
	**                (wirkt beim Kuboid spinnenartig) in gedeckter Farbe ...
	*/
/* 	for (owgp = wg, pos=0; owgp; owgp = owgp->nextwgp, pos++) { */
/* 	  if (pos % 2 != 0) { */
/* 	    continue; */
/* 	  } */

/*           wErr(glPushMatrix()); */

/* 	  if (options.axistrans == 2) { */
/* 	    wErr(glRotatef (GWINKEL(owgp->wgp.w2), 1., 0., 0.)); */
/* 	    wErr(glRotatef (GWINKEL(owgp->wgp.w),  0., 1., 0.)); */
/* 	  } */
/* 	  if (options.axistrans == 1) { */
/* 	    wErr(glRotatef (GWINKEL(owgp->wgp.w),  0., 1., 0.)); */
/* 	  } */

/* 	  wErr(glColor3f (0.4,0.4,0.2)); */
/* 	  wErr(glClearColor (0.7, 0.7, 0.0, 0.0)); */
/* 	  wErr(glLineWidth (1.)); */
/* 	  wErr(glBegin (GL_LINES)); */

/* 	  if (options.oloid) { */
/*             wErr(glVertex3f (VON(3)));  */
/*             wErr(glVertex3f (VON(17))); */
/* 	  } */
/* 	  if (options.kuboid) { */
/*             wErr(glVertex3f (VON(3))); */
/*             wErr(glVertex3f (VON(7))); */
/*             wErr(glVertex3f (VON(7))); */
/*             wErr(glVertex3f (VON(5))); */
/* 	  } */
/* 	  wErr(glEnd ()); */

/*           wErr(glPopMatrix()); */
/* 	} */


        /*
	** Verbindungslinien zwischen Punkten zweier Folgepositionen
	*/

	/* wgw1 zeigt das Oloid aber auch das gebogene Kuboid (Polkuboid) */
	/* wgw2 zeigt das Kuboid (Umstuelpungshalle) und das Oloid-Huetchen */
	if (options.axistrans == 2) {
	  actualwgp = wgw1;
	  lastactualwgp = lastw1;
	}
	else if (options.axistrans == 1) {
	  actualwgp = wgw2;
	  lastactualwgp = lastw2;
	}
	else {
	  /* keine Transformation */
	  actualwgp = wg;
	  lastactualwgp = last;
	}
	for (owgp = actualwgp, pos=0; owgp; owgp = owgp->nextwgp, pos++) {
	  WG *prevwgp;
	  prevwgp = (owgp == actualwgp) ? lastactualwgp : owgp->prevwgp;

/* 	  wErr(glBegin (GL_LINES)); */
/* 	  if (options.oloid) { */
/* 	      wErr(glVertex3f (PREVVON(3))); wErr(glColor3f (PREVVON(3))); */
/* 	      wErr(glVertex3f (VON(3))); wErr(glColor3f (VON(3))); */
/* 	      wErr(glVertex3f (VON(3))); wErr(glColor3f (VON(3))); */
/* 	      wErr(glVertex3f (PREVVON(17))); wErr(glColor3f (PREVVON(7))); */
/* 	      wErr(glVertex3f (PREVVON(17))); wErr(glColor3f (PREVVON(7))); */
/* 	      wErr(glVertex3f (PREVVON(3))); wErr(glColor3f (PREVVON(3))); */
/* 	      wErr(glVertex3f (VON(3))); wErr(glColor3f (VON(3))); */
/* 	      wErr(glVertex3f (VON(17))); wErr(glColor3f (VON(7))); */
/* 	      wErr(glVertex3f (VON(17))); wErr(glColor3f (VON(7))); */
/* 	      wErr(glVertex3f (PREVVON(17))); wErr(glColor3f (PREVVON(7))); */
/* 	      wErr(glVertex3f (PREVVON(17))); wErr(glColor3f (PREVVON(7))); */
/* 	      wErr(glVertex3f (VON(3))); wErr(glColor3f (VON(3))); */
/* 	  } */
/* 	  else { */
/* 	      wErr(glVertex3f (PREVVON(3))); wErr(glColor3f (PREVVON(3))); */
/* 	      wErr(glVertex3f (VON(3))); wErr(glColor3f (VON(3))); */
/* 	      wErr(glVertex3f (VON(3))); wErr(glColor3f (VON(3))); */
/* 	      wErr(glVertex3f (VON(7))); wErr(glColor3f (VON(7))); */
/* 	      wErr(glVertex3f (VON(7))); wErr(glColor3f (VON(7))); */
/* 	      wErr(glVertex3f (PREVVON(7))); wErr(glColor3f (PREVVON(7))); */
/* 	      wErr(glVertex3f (PREVVON(7))); wErr(glColor3f (PREVVON(7))); */
/* 	      wErr(glVertex3f (PREVVON(3))); wErr(glColor3f (PREVVON(3))); */
/* 	      wErr(glVertex3f (VON(7))); wErr(glColor3f (VON(7))); */
/* 	      wErr(glVertex3f (VON(5))); wErr(glColor3f (VON(5))); */
/* 	      wErr(glVertex3f (VON(5))); wErr(glColor3f (VON(5))); */
/* 	      wErr(glVertex3f (PREVVON(5))); wErr(glColor3f (PREVVON(5))); */
/* 	      wErr(glVertex3f (PREVVON(5))); wErr(glColor3f (PREVVON(5))); */
/* 	      wErr(glVertex3f (PREVVON(7))); wErr(glColor3f (PREVVON(7))); */
/* 	      wErr(glVertex3f (PREVVON(7))); wErr(glColor3f (PREVVON(7))); */
/* 	      wErr(glVertex3f (VON(7))); wErr(glColor3f (VON(7))); */
/* 	  } */
/* 	  wErr(glEnd ()); */

// keine Flaechen         /* die Flaechendarstellung mit GL_QUADS */
// keine Flaechen 	  wErr(glBegin (GL_QUADS));
// keine Flaechen 	  if (options.oloid) {
// keine Flaechen 	      wErr(glVertex3f (PREVVON(3))); wErr(glColor3f (PREVVON(3)));
// keine Flaechen 	      wErr(glVertex3f (VON(3))); wErr(glColor3f (VON(3)));
// keine Flaechen 	      wErr(glVertex3f (VON(3))); wErr(glColor3f (VON(3)));
// keine Flaechen 	      wErr(glVertex3f (VON(17))); wErr(glColor3f (VON(7)));
// keine Flaechen 	      wErr(glVertex3f (VON(17))); wErr(glColor3f (VON(7)));
// keine Flaechen 	      wErr(glVertex3f (PREVVON(17))); wErr(glColor3f (PREVVON(7)));
// keine Flaechen 	      wErr(glVertex3f (PREVVON(17))); wErr(glColor3f (PREVVON(7)));
// keine Flaechen 	      wErr(glVertex3f (PREVVON(3))); wErr(glColor3f (PREVVON(3)));
// keine Flaechen 	  }
// keine Flaechen 	  else {
// keine Flaechen 	      wErr(glVertex3f (PREVVON(3))); wErr(glColor3f (PREVVON(3)));
// keine Flaechen 	      wErr(glVertex3f (VON(3))); wErr(glColor3f (VON(3)));
// keine Flaechen 	      wErr(glVertex3f (VON(3))); wErr(glColor3f (VON(3)));
// keine Flaechen 	      wErr(glVertex3f (VON(7))); wErr(glColor3f (VON(7)));
// keine Flaechen 	      wErr(glVertex3f (VON(7))); wErr(glColor3f (VON(7)));
// keine Flaechen 	      wErr(glVertex3f (PREVVON(7))); wErr(glColor3f (PREVVON(7)));
// keine Flaechen 	      wErr(glVertex3f (PREVVON(7))); wErr(glColor3f (PREVVON(7)));
// keine Flaechen 	      wErr(glVertex3f (PREVVON(3))); wErr(glColor3f (PREVVON(3)));
// keine Flaechen 	      wErr(glVertex3f (VON(7))); wErr(glColor3f (VON(7)));
// keine Flaechen 	      wErr(glVertex3f (VON(5))); wErr(glColor3f (VON(5)));
// keine Flaechen 	      wErr(glVertex3f (VON(5))); wErr(glColor3f (VON(5)));
// keine Flaechen 	      wErr(glVertex3f (PREVVON(5))); wErr(glColor3f (PREVVON(5)));
// keine Flaechen 	      wErr(glVertex3f (PREVVON(5))); wErr(glColor3f (PREVVON(5)));
// keine Flaechen 	      wErr(glVertex3f (PREVVON(7))); wErr(glColor3f (PREVVON(7)));
// keine Flaechen 	      wErr(glVertex3f (PREVVON(7))); wErr(glColor3f (PREVVON(7)));
// keine Flaechen 	      wErr(glVertex3f (VON(7))); wErr(glColor3f (VON(7)));
// keine Flaechen 	  }
// keine Flaechen 	  wErr(glEnd ());


	  /* die Flaechendarstellung mit GL_TRINAGLE_FAN
	  ** (P vor dem Punkt bedeutet Punkt der vorherigen {previous} Position)
	  ** Oloid: eine Flaeche verbindet Punkte 3-17-P17-P3
	  **
	  ** Kuboid: eine Flaeche verbindet Punkte 7-3-P3-P7-P5-5
	  */
	  wErr(glBegin (GL_TRIANGLE_FAN));
	  if (options.oloid || options.beide) {
	      wErr(glVertex3f (VON(3))); wErr(glColor3f (VON(3)));
	      wErr(glVertex3f (VON(17))); wErr(glColor3f (VON(7)));
	      wErr(glVertex3f (PREVVON(17))); wErr(glColor3f (PREVVON(17)));
	      wErr(glVertex3f (PREVVON(3))); wErr(glColor3f (PREVVON(3)));
	  }
	  if (options.kuboid || options.beide) {
	      wErr(glVertex3f (VON(7))); wErr(glColor3f (VON(7)));
	      wErr(glVertex3f (VON(3))); wErr(glColor3f (VON(3)));
	      wErr(glVertex3f (PREVVON(3))); wErr(glColor3f (PREVVON(3)));
	      wErr(glVertex3f (PREVVON(7))); wErr(glColor3f (PREVVON(7)));
	      wErr(glVertex3f (PREVVON(5))); wErr(glColor3f (PREVVON(5)));
	      wErr(glVertex3f (VON(5))); wErr(glColor3f (VON(5)));
	  }
	  wErr(glEnd ());

	}
}

void
wCreateTimeform2 () {
	WG *owgp;
	WG *actualwgp;
	WG *lastactualwgp;

	/*
	** Oloid/Kuboid erst die Verbindungslinien der Endpunkte dunkel,
	** dann Endpunkte selbst hell
	*/
	/* wgw1 zeigt das Oloid aber auch das gebogene Kuboid (Polkuboid) */
	/* wgw2 zeigt das Kuboid (Umstuelpungshalle) und das Oloid-Huetchen */
	if (options.axistrans == 2) {
	  actualwgp = wgw1;
	  lastactualwgp = lastw1;
	}
	else if (options.axistrans == 1) {
	  actualwgp = wgw2;
	  lastactualwgp = lastw2;
	}
	else {
	  /* keine Transformation */
	  actualwgp = wg;
	  lastactualwgp = last;
	}
	for (owgp = actualwgp; owgp; owgp = owgp->nextwgp) {
	  WG *prevwgp;
	  prevwgp = (owgp == actualwgp) ? lastactualwgp : owgp->prevwgp;

	  /* Verbindungslinien dunkel */
	  wErr(glColor3f (0.4,0.4,0.2));
	  wErr(glClearColor (0.7, 0.7, 0.0, 0.0));
	  wErr(glLineWidth (1.));
	  wErr(glBegin (GL_LINES));

	  if (options.oloid || options.beide) {
	    /* das Oloid, das Oloid-Htchen oder Konoid-aehnliche Figur 3-17
	    ** je nachdem, welche Quelle oben gewaehlt wurde (welche Achsentransf.)
	    */
            wErr(glVertex3f (PREVVON(3)));
            wErr(glVertex3f (VON(3)));
            wErr(glVertex3f (PREVVON(17)));
            wErr(glVertex3f (VON(17)));
            wErr(glVertex3f (PREVVON(22)));
            wErr(glVertex3f (VON(22)));
	    /* verkehrtes Htchen 7-13 */
/*             wErr(glVertex3f (PREVVON(7))); */
/*             wErr(glVertex3f (VON(7))); */
/*             wErr(glVertex3f (PREVVON(13))); */
/*             wErr(glVertex3f (VON(13))); */
	    /* ein grosser Kreis 1-5 */
/*             wErr(glVertex3f (PREVVON(1))); */
/*             wErr(glVertex3f (VON(1))); */
/*             wErr(glVertex3f (PREVVON(5))); */
/*             wErr(glVertex3f (VON(5))); */
	  }
	  if (options.kuboid || options.beide) {
            wErr(glVertex3f (PREVVON(3)));
            wErr(glVertex3f (VON(3)));
            wErr(glVertex3f (PREVVON(7)));
            wErr(glVertex3f (VON(7)));
            wErr(glVertex3f (PREVVON(5)));
            wErr(glVertex3f (VON(5)));
            wErr(glVertex3f (PREVVON(22)));
            wErr(glVertex3f (VON(22)));
	  }
	  wErr(glEnd ());



	  /* Endpunkte hell */
	  wErr(glColor3f (1.,1.,1.));
	  wErr(glClearColor (0.7, 0.7, 0.2, 0.0));
	  wErr(glPointSize (2.));
	  wErr(glBegin (GL_POINTS));

	  if (options.oloid || options.beide) {
            wErr(glVertex3f (VON(3))); 
            wErr(glVertex3f (VON(17)));
            wErr(glVertex3f (VON(22)));
	  }
	  if (options.kuboid || options.beide) {
            wErr(glVertex3f (VON(3)));
            wErr(glVertex3f (VON(7)));
            wErr(glVertex3f (VON(5)));
            wErr(glVertex3f (VON(22)));
	  }

	  wErr(glEnd ());
	}
}

void
wCreateVirt () {
        wErr(glColor3f (1.0, 1.0, 1.0));
        wErr(glClearColor (0.7, 0.7, 0.0, 0.0));
        wErr(glLineWidth (1.0));

	wErr(glBegin (GL_LINES));
	  wErr(glVertex3f (VON(4)));  wErr(glVertex3f (NACH(9)));
	  wErr(glVertex3f (VON(6)));  wErr(glVertex3f (NACH(9)));
	  wErr(glVertex3f (VON(14))); wErr(glVertex3f (NACH(9)));

	  wErr(glVertex3f (VON(2)));  wErr(glVertex3f (NACH(10)));
	  wErr(glVertex3f (VON(8)));  wErr(glVertex3f (NACH(10)));
	  wErr(glVertex3f (VON(18))); wErr(glVertex3f (NACH(10)));
	wErr(glEnd ());
}

void
wCreateAxis () {
  GLfloat params[4];

        wErr(glColor3f (1.0, 1.0, 1.0));
        wErr(glClearColor (0.7, 0.7, 0.0, 0.0));
        wErr(glLineWidth (1.0));

        wErr(glBegin (GL_LINES));
          wErr(glVertex3f (VON(1))); wErr(glVertex3f (NACH(19)));
          wErr(glVertex3f (VON(1))); wErr(glVertex3f (NACH(20)));
          wErr(glVertex3f (VON(1))); wErr(glVertex3f (NACH(21)));
        wErr(glEnd ());

  wErr(glGetFloatv(GL_CURRENT_COLOR, params)); 
  wErr(glColor3f (0., 0., 0.)); /* Labels immer in Schwarz */
        wErr(glRasterPos3f (VON(19))); wDrawString ("X");
        wErr(glRasterPos3f (VON(20))); wDrawString ("Y");
        wErr(glRasterPos3f (VON(21))); wDrawString ("Z");
  wErr(glColor4fv (params));
}

void
wCreateDiagonals () {
       /*
       ** Wuerfeldiagonale zwischen den beiden Drehpunkten P1-P5
       ** sowie die beiden anderen konstanten Diagonalen
       ** vielleicht auch sinnvoll Verbindung zwischen I uen E
       */
       wErr(glLineWidth (1.0));
       wErr(glColor3f (.8,.8,.3));
       wErr(glClearColor (0.7, 0.7, 0.0, 0.0));
       wErr(glBegin (GL_LINES));
         wErr(glVertex3f (VON(1)));
         wErr(glVertex3f (NACH(5)));
         wErr(glVertex3f (VON(3)));
         wErr(glVertex3f (NACH(17)));
         wErr(glVertex3f (VON(7)));
         wErr(glVertex3f (NACH(13)));
         wErr(glVertex3f (VON(9)));
         wErr(glVertex3f (NACH(10)));
       wErr(glEnd ());
}

void
wCreateCenter () {
       /*
       ** Mittelpunkt des Wuerfelguertels
       */

       wErr(glPointSize(4.));
       wErr(glColor3f (1., 1., 1.));
       wErr(glClearColor (0.7, 0.7, 0.0, 0.0));
       wErr(glBegin (GL_POINTS));
         wErr(glVertex3f (VON(22)));
       wErr(glEnd ());
}


/*
** diplayMatrix gibt die 16 hintereinander liegenden floats in einer column-major
** Reihenfolge aus. m ist ein beliebiger erklaerender Text.
*/
void
wDisplayMatrix (char *m, GLfloat *f) {
  int i;

  fprintf (stderr, "%s\n", m);
  for (i=0; i<16; i+=4) {
    fprintf (stderr, "  %5.2f %5.2f %5.2f %5.2f\n", f[i], f[i+4], f[i+8], f[i+12]);
  }

}

void
wDisplayGlMatrices () {
    GLfloat modelviewMatrix[16];
    GLfloat projectionMatrix[16];
    char text[100];

/*     wErr(glGetFloatv (GL_MODELVIEW_MATRIX, &mvMatrix[0])); */
/*     wDisplayMatrix ("mvMatrix start", &mvMatrix[0]); */
/*     wErr(glGetFloatv (GL_PROJECTION_MATRIX, &pvMatrix[0])); */
/*     wDisplayMatrix ("pvMatrix start", &pvMatrix[0]); */

    sprintf (text, "modelviewMatrix in die x/y Achse bei w=%6.2f, w2=%6.2f",
	     GWINKEL(owgp->wgp.w), GWINKEL(owgp->wgp.w2));
    wErr(glGetFloatv (GL_MODELVIEW_MATRIX, &modelviewMatrix[0]));
    wDisplayMatrix (text, &modelviewMatrix[0]);
    sprintf (text, "projectionMatrix in die x/y Achse bei w=%6.2f, w2=%6.2f",
	     GWINKEL(owgp->wgp.w), GWINKEL(owgp->wgp.w2));
    wErr(glGetFloatv (GL_PROJECTION_MATRIX, &projectionMatrix[0]));
    wDisplayMatrix (text, &projectionMatrix[0]);
}
